#include "stdafx.h"
#include "Ilpilbox.h"
#include "Interface.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// PIL-Box special init command frames:
#define TDIS	0x494						// TDIS
#define COFI	0x495						// COFF with IDY, firmware >= v1.6
#define CON		0x496						// Controller ON
#define COFF	0x497						// Controller OFF
#define SSRQ	0x49C						// Set Service Request
#define CSRQ	0x49D						// Clear Service Request

extern CIlpilboxApp theApp;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CInterface::CInterface(LPCTSTR lpszComPort, DWORD dwBaudRate, bool bIDY, bool bEnableRFC, CWnd* pParent)
	: m_dwBaudRate(dwBaudRate), m_bIDY(bIDY), m_bEnableRFC(bEnableRFC), m_bUSE8BITS(true), m_pParent(pParent)
{
	m_hWorkerThread = NULL;

	m_hComm = NULL;							// COM port not opened

	m_bConMode = false;						// no switch to CON mode
	m_eState = LS_IDLE;						// no transfer

	m_bRecvEn = true;						// receiver enabled

	// event data send to PilBox
	m_hEvent = CreateEvent(NULL,FALSE,FALSE,NULL);

	// event to kill timeout
	m_hEvTimeout = CreateEvent(NULL,FALSE,FALSE,NULL);

	m_hComm = CreateFile(CString(_T("\\\\.\\")) + lpszComPort,
						 GENERIC_READ | GENERIC_WRITE,
						 0,
						 NULL,
						 OPEN_EXISTING,
						 FILE_ATTRIBUTE_NORMAL,
						 NULL);

	if (m_hComm == INVALID_HANDLE_VALUE)
	{
		m_hComm = NULL;
		m_pParent->MessageBox(_T("Error: COM port not available."),_T("ILPilbox"),MB_OK | MB_ICONEXCLAMATION);
		return;
	}

	if (!SetPILBox(COFF,true))				// init the PIL-Box in control-off mode
	{
		CloseHandle(m_hComm);				// close port
		m_hComm = NULL;						// error initializing

		m_pParent->MessageBox(_T("Error: no response from PILBox."),_T("ILPilbox"),MB_OK | MB_ICONEXCLAMATION);
	}
}

CInterface::~CInterface()
{
	if (m_hComm)
	{
		VERIFY(SetPILBox(TDIS));			// stop the PIL-Box
		CloseHandle(m_hComm);				// close port
	}
	CloseHandle(m_hEvent);					// close event
	CloseHandle(m_hEvTimeout);
}

VOID CInterface::InitComm(DWORD dwBaudrate)
{
	if (m_hComm != NULL)
	{
		DCB dcb;

		ZeroMemory(&dcb,sizeof(dcb));
		dcb.DCBlength = sizeof(dcb);
		dcb.BaudRate = dwBaudrate;
		dcb.fBinary = TRUE;
		dcb.fParity = TRUE;
		dcb.fOutxCtsFlow = FALSE;
		dcb.fOutxDsrFlow = FALSE;
		dcb.fDtrControl = DTR_CONTROL_ENABLE; // enable DTR line for RP2040 compatibility
		dcb.fDsrSensitivity = FALSE;
		dcb.fOutX = FALSE;
		dcb.fErrorChar = FALSE;
		dcb.fNull = FALSE;
		dcb.fRtsControl = RTS_CONTROL_DISABLE;
		dcb.fAbortOnError = FALSE;
		dcb.ByteSize = 8;
		dcb.Parity = NOPARITY;				// no parity check
		dcb.StopBits = ONESTOPBIT;

		// com port settings
		SetCommState(m_hComm,&dcb);
	}
	return;
}

DWORD CInterface::GetTimeout(WORD wFrame)
{
	DWORD dwTimeout;

	// wait for returned frame
	// timeouts: 10s for DOE/RDY, 3s for CMD, 1s for IDY 
	do
	{
		if (wFrame < 0x400)					// DOE
		{
			dwTimeout = 10000;
			break;
		}
		if ((wFrame & 0x7F4) == 0x494)		// special PILBox frames 494-497 and 49c-49f
		{
			dwTimeout = 500;
			break;
		}
		if (wFrame < 0x500)					// CMD
		{
			dwTimeout = 3000;
			break;
		}
		if (wFrame < 0x600)					// RDY
		{
			dwTimeout = 10000;
			break;
		}
		dwTimeout = 1000;					// IDY
	}
	while (FALSE);

	return dwTimeout;
}

// --------------------------------
// SetPILBox()
// set a PIL Box command
// --------------------------------

bool CInterface::SetPILBox(WORD wCmd,bool bInit)
{
	bool bSucc;

	COMMTIMEOUTS CommCmdTimeouts  = { 0L, 0L, 50L, 0L, 0L };
	COMMTIMEOUTS CommPollTimeouts = { 0L, 0L,  2L, 0L, 0L };

	// timeouts for PILBox communication
	SetCommTimeouts(m_hComm,&CommCmdTimeouts);

	if (bInit)								// initialize with baud rate search
	{
		if (m_dwBaudRate == 0)				// auto baudrate
		{
			// list with possible baudrates
			CONST DWORD dwBpsList[] = { 460800, 230400, 115200, 9600 };

			UINT i = 0;						// begin with first baudrate

			do
			{
				m_dwBaudRate = dwBpsList[i++]; // get baudrate

				InitComm(m_dwBaudRate);		// set baudrate and timeout
				PurgeComm(m_hComm,PURGE_RXCLEAR); // clear receive buffer
				bSucc = PILBoxCmd(wCmd);	// init PILBox
			}
			while (!bSucc && i < ARRAYSIZEOF(dwBpsList));
		}
		else								// manual baudrate
		{
			InitComm(m_dwBaudRate);			// set baudrate and timeout
			bSucc = PILBoxCmd(wCmd);		// init PILBox
		}
	}
	else
	{
		bSucc = PILBoxCmd(wCmd);
	}

	if (m_bIDY && bSucc && wCmd == COFF)	// enable control-off mode with IDY
	{
		bSucc = PILBoxCmd(COFI);			// init the PIL-Box in control-off mode with IDY
	}

	m_wMode = wCmd;							// save the last setting

	// timeouts for PILBox polling
	SetCommTimeouts(m_hComm,&CommPollTimeouts);
	return bSucc;
}

// --------------------------------
// PILBoxCmd()
// execute a PIL Box command
// --------------------------------

bool CInterface::PILBoxCmd(WORD wCmd)
{
	BYTE  byRx;
	DWORD dwBytesRead;

	m_byLastHigh = 0;
	SendFrame(wCmd);

	if (ReadFile(m_hComm,&byRx,sizeof(byRx),&dwBytesRead,NULL) == FALSE)
		dwBytesRead = 0L;

	// true on success
	return dwBytesRead != 0 && ((byRx & 0x3F) == (wCmd & 0x3F));
}

// --------------------------------
// SendFrame()
// send a IL frame to the PILBox
// --------------------------------

VOID CInterface::SendFrame(WORD wFrame)
{
	BYTE  byLow,byHigh;						// low/high bytes to serial
	BYTE  byBuf[2];
	DWORD n,dwWritten;

	// build the low and high parts
	if (m_bUSE8BITS == false)
	{
		// use 7-bit characters for maximum compatibility
		byHigh = ((wFrame >> 6) & 0x1F) | 0x20;
		byLow  = (wFrame & 0x3F) | 0x40;
	}
	else
	{
		// 8-bit format for optimum speed
		byHigh = ((wFrame >> 6) & 0x1E) | 0x20;
		byLow  = (wFrame & 0x7F) | 0x80;
	}
	if (byHigh != m_byLastHigh)
	{
		// send high part if different from last one
		m_byLastHigh = byHigh;
		byBuf[0] = byHigh;
		byBuf[1] = byLow;
		n = 2;
	}
	else
	{
		// otherwise send only low part
		byBuf[0] = byLow;
		n = 1;
	}
	WriteFile(m_hComm,byBuf,n,&dwWritten,NULL);
	ASSERT(n == dwWritten);
	return;
}

// --------------------------------
// PILBox()
// manage the PILBox
// (based on Emu41 ilext2 module)
// called at each received byte
// --------------------------------

VOID CInterface::PILBox(BYTE byVal)
{
	WORD  wFrame;							// IL frame
	DWORD dwTimeout;

	TRACE(_T("PILBox: %02X\n"),byVal);

	if ((byVal & 0xC0) == 0)				// fetched high byte, ACK or garbage
	{
		if ((byVal & 0x20) != 0)			// fetched high byte
		{
			m_byLastHigh = byVal;			// high byte, save it

			if (m_dwBaudRate == 9600)		// only 9600 baud operation need ACK
			{
				DWORD dwWritten;

				byVal = 13;					// acknowledge
				WriteFile(m_hComm,&byVal,sizeof(byVal),&dwWritten,NULL);
				ASSERT(1 == dwWritten);
				TRACE(_T("PILBox acknowledge\n"));
			}
		}
	}
	else
	{
		// low byte, build frame according to format
		if (m_bUSE8BITS = ((byVal & 0x80) != 0))
			wFrame = ((m_byLastHigh & 0x1E) << 6) | (byVal & 0x7F);
		else
			wFrame = ((m_byLastHigh & 0x1F) << 6) | (byVal & 0x3F);

		ResetEvent(m_hEvent);				// clear event of last SendFrame()

		if (m_eState == LS_IDLE)
		{
			m_eState = LS_PILBOX;			// initiator is PILBox
		}

		// transmit IL frame data to virtual HP-IL loop
		TRACE(_T("PILBox frame data to virtual HP-IL loop : %02X\n"),wFrame);
		theApp.m_pTcpIpObj->SendFrame(wFrame);

		switch (m_eState)
		{
		case LS_PILBOX:
			m_eState = LS_PILBOX;			// initiator is PILBox
			dwTimeout = GetTimeout(wFrame);	// get frame specific timeout

			// wait for finishing the virtual HP-IL loop with a last
			// SendFrame() to the Pilbox before continue reading
			if (WaitForSingleObject(m_hEvent,dwTimeout) != WAIT_OBJECT_0)
			{
				m_eState = LS_IDLE;

				// loop broken, clear remaining Pilbox data
				PurgeComm(m_hComm,PURGE_RXCLEAR);
			}
			break;
		case LS_TCPIP:
			m_eState = LS_IDLE;
			SetEvent(m_hEvent);				// continue reading next frame from Pilbox
			break;
		case LS_IDLE:
			break;
		default: ASSERT(FALSE);
		}
	}
	return;
}

// --------------------------------
// SendFrameCtrl()
// send a IL frame in the correct
// mode to the PILBox
// --------------------------------

VOID CInterface::SendFrameCtrl(WORD wFrame)
{
	static WORD wCmdFrame = 0;

	DWORD dwTimeout;

	if (   m_eState == LS_IDLE				// transfer initiator is TCP/IP
		|| m_eState == LS_TCPIP)			// or still in TCP/IP mode because PILBox got no answer from virtual loop
	{
		m_dwReftime = GetTickCount();		// last time of writing

		if (m_wMode != CON)					// PILBox is not controller
		{
			m_bConMode = true;				// set PILBox into controller mode
			while (m_bConMode || m_wMode != CON) // wait PILBox set into controller mode
				Sleep(0);
		}

		m_bRecvEn = false;					// receiver disabled

		dwTimeout = GetTimeout(wFrame);		// get frame specific timeout
	}
	else									// transfer initiator is PILBox
	{
		if (m_bEnableRFC)					// RFC frame enabled?
		{
			if ((wFrame & 0x700) == 0x400)	// received a CMD frame from the virtual loop
			{
				wCmdFrame = wFrame;			// remember the CMD frame

				// send RFC frame to the virtual loop
				theApp.m_pTcpIpObj->SendFrame(0x500);
				return;
			}

			// RFC frame and saved CMD frame is valid
			if (wFrame == 0x500 && (wCmdFrame & 0x700) == 0x400)
			{
				wFrame = wCmdFrame;			// restore original CMD frame
			}

			wCmdFrame = 0;					// no CMD frame
		}
	}

	SendFrame(wFrame);						// send the frame to the PILBox
	m_bRecvEn = true;						// receiver enabled
	SetEvent(m_hEvTimeout);					// kill timeout

	switch (m_eState)
	{
	case LS_IDLE:
		m_eState = LS_TCPIP;
		// no break
	case LS_TCPIP:
		// wait for PILBox send data to TCP/IP
		WaitForSingleObject(m_hEvent,dwTimeout);
		m_bRecvEn = false;					// receiver disabled
		break;
	case LS_PILBOX:
		m_eState = LS_IDLE;
		SetEvent(m_hEvent);					// continue reading next frame from PILBox
		break;
	default: ASSERT(FALSE);
	}
	return;
}

UINT CInterface::ComThreadProc(LPVOID pParam)
{
	CInterface *p = reinterpret_cast<CInterface *>(pParam);

	BYTE  byRx;
	DWORD dwBytesRead;

	ASSERT(p != NULL);

	while (p->m_bRunning)
	{
		// PILBox in device mode and CON mode switch requested
		if (p->m_wMode == COFF && p->m_bConMode)
		{
			COMMTIMEOUTS CommTimeouts = { MAXDWORD, 0L, 0L, 0L, 0L };

			TRACE(_T("**CON mode**\n"));
			VERIFY(p->SetPILBox(CON));		// switch PILBox into controller mode

			// timeouts for controller mode
			SetCommTimeouts(p->m_hComm,&CommTimeouts);

			p->m_bConMode = false;			// switch to CON mode accepted
		}

		// PILBox in controller mode and 15s timeout elapsed
		if (p->m_wMode == CON && (GetTickCount() - p->m_dwReftime) > 15000)
		{
			COMMTIMEOUTS CommTimeouts = { 0L, 0L, 2L, 0L, 0L };

			p->m_bRecvEn = true;			// receiver enabled
			SetEvent(p->m_hEvTimeout);		// kill timeout

			TRACE(_T("**COFF mode**\n"));
			VERIFY(p->SetPILBox(COFF));		// switch PILBox back into device mode

			// timeouts for device mode
			SetCommTimeouts(p->m_hComm,&CommTimeouts);
		}

		if (p->m_bRecvEn)					// receiver enabled
		{
			// normal receive from PILBox
			if (ReadFile(p->m_hComm,&byRx,sizeof(byRx),&dwBytesRead,NULL) == FALSE)
				dwBytesRead = 0L;

			if (dwBytesRead > 0)			// something received
			{
				p->PILBox(byRx);			// handle command
			}
		}
		else								// receiver disabled
		{
			// CON mode idle state
			WaitForSingleObject(p->m_hEvTimeout,50);
		}
	}
	return 0;
}

bool CInterface::Connect()
{
	if (m_hComm != NULL)
	{
		// create server thread
		DWORD dwThreadID;

		m_bRunning = true;
		m_hWorkerThread = CreateThread(NULL,
									   0,
									   (LPTHREAD_START_ROUTINE) ComThreadProc,
									   (LPVOID) this,
									   0,
									   &dwThreadID);
		ASSERT(m_hWorkerThread);
	}
	return m_hWorkerThread != NULL;
}

VOID CInterface::Disconnect()
{
	if (m_hWorkerThread)					// server running
	{
		m_bRunning = false;					// stop Pilbox interface thread

		WaitForSingleObject(m_hWorkerThread,INFINITE);
		CloseHandle(m_hWorkerThread);
		m_hWorkerThread = NULL;
	}
	return;
}
